use crate::bkz::bkz_reduce;
use crate::matrix::Matrix;
use crate::utils::abs;
use lll_rs::l2::bigl2;
use rug::Integer;

pub fn agcd(ciphertexts: Vec<Integer>, noise_bits: usize, algorithm: u8) -> Integer {
    // 1. Build lattice matrix basis
    let basis_matrix = Matrix::new_lattice(noise_bits, ciphertexts.clone()).unwrap();

    // 2. reduce with LLL
    let mut lll_matrix = basis_matrix.to_lll_matrix();
    println!("basis: {:?}", lll_matrix);
    match algorithm {
        0u8 => bigl2::lattice_reduce(&mut lll_matrix, 0.51, 0.75),
        1u8 => bkz_reduce(&mut lll_matrix, 16, 0.75, 0.75, 10),
        _ => panic!(),
    }
    println!("basis after reduction: {:?}", lll_matrix);

    // 3. Extract shortest vector
    let shortest_vector = &lll_matrix[0];
    println!("Shortest vector: {:?}", shortest_vector);

    // 4. q0 candidate
    let q0 = &shortest_vector[0] / (Integer::from(1) << (noise_bits + 1));
    println!("q0: {}", q0);

    if q0 == 0 {
        println!("Couldn't recover p.");
        return Integer::from(0);
    }

    // 5. Find p
    // compute r0 = x0 (mod q0)
    // and p = (x0 − r0)/q0.
    let x0 = &ciphertexts[0];
    println!("x0: {}", x0);
    let r0 = x0 % q0.clone();
    println!("r0: {}", r0);
    let p_guess = abs((x0 - r0) / q0);

    println!("Recovered p: {}", p_guess);
    println!(
        "Approximate GCD with noise_bits={}: {}",
        noise_bits, p_guess
    );

    p_guess
}
